#!/usr/bin/perl

######################################################################
package Cognition::uF::hCard;
######################################################################

use XML::LibXML qw(:all);
use strict;
use CGI::Util;
use Digest::SHA1 qw(sha1_hex);

BEGIN
{
	our $VERSION = 0.1;
	our @ISA = qw(Cognition::uF);
}


sub MAX_INCLUDE_ITERATIONS { 6; }


sub parse_all
{
	my $page   = shift;
	my $within = shift || $page->{DOM};
	my @nodes = Cognition::HTMLParser::Utils::searchClass('vcard', $within);
	my @rv;
	
	foreach my $card (@nodes)
	{
		my $hcard = parse($page, $card);
		push @rv, $hcard;
	}
	
	return @rv;
} #/sub parse_all


sub uri
{
	my $this = shift;
	my $all  = shift;
	my @rv;

	if (length $this->{uid})
	{
		push @rv, CGI::Util::escape($this->{uid});
	}
	if (length $this->{_id})
	{
		push @rv, $this->{_page}->abs_url.'#'.$this->{_id};
	}
	if (length $this->{fn})
	{
		push @rv, 'urn:x-hcard:' . CGI::Util::escape($this->{fn});
	}
	push @rv, Cognition::HTMLParser::Utils::XPointer($this->{_dom}, $this->{_page}->abs_url);
	
	return $rv[0] unless (defined $all);
	while ($all) { shift @rv; $all--; }
	return @rv;
} #/sub uri


sub rdf_subject
{
	my $this    = shift;
	my $ns      = 'urn:ietf:rfc:2426#';

	# Create RDF subject
	my $subject = Cognition::RDFModel::Subject->new($this->uri);
	$subject->addObject($this);
	$subject->setDom($this->{_dom});
	
	$subject->addVal('http://www.w3.org/1999/02/22-rdf-syntax-ns#type',
		Cognition::RDFModel::Value->new($ns.'vcard', 1));
	
	# Patterns for simple data.
	my @_SINGLES  = qw(fn bday tz sort-string class family-name given-name
	                   additional-name honorific-prefix honorific-suffix n);
	my @_SINGLESR = qw(uid);
	my @_MULTIS   = qw(nickname label title role key mailer category note rev);
	my @_MULTISR  = qw(url photo sound logo);
	my @_OBJECTS  = qw(category adr geo agent);
	
	# Properties which exist as singles.
	foreach my $property (@_SINGLES)
	{
		next unless (length $this->{$property});
		my $val = Cognition::RDFModel::Value->new($this->{$property});
		$subject->addVal($ns.$property, $val);
	}

	# Resource-type properties which exist as singles.
	foreach my $property (@_SINGLESR)
	{
		next unless (length $this->{$property});
		my $val = Cognition::RDFModel::Value->new($this->{$property}, 1);
		$subject->addVal($ns.$property, $val);
	}
		
	# Properties which exist as multiple values.
	foreach my $property (@_MULTIS)
	{
		foreach my $V (@{ $this->{$property} })
		{
			next unless (length $V);
			my $val = Cognition::RDFModel::Value->new($V);
			$subject->addVal($ns.$property, $val);
		}
	}
	
	# Resource-type properties which exist as multiple values.
	foreach my $property (@_MULTISR)
	{
		foreach my $V (@{ $this->{$property} })
		{
			next unless (length $V);
			my $val = Cognition::RDFModel::Value->new($V, 1);
			$subject->addVal($ns.$property, $val);
		}
	}

	# E-mail
	foreach my $V (@{ $this->{email} })
	{
		my $v = 'mailto:'.$V->{value};
		my $val = Cognition::RDFModel::Value->new($v, 1);
		$subject->addVal($ns.'email', $val);
	}

	# Telephone
	foreach my $V (@{ $this->{tel} })
	{
		my $v = $V->{value};
		if ($v =~ /x/)
		{
			$v =~ s/x/;ext=/;
		}
		if ($v !~ /^\+/)
		{
			$v .= ';phone-context=unknown-context.invalid';
		}
		$v = "tel:$v";
		my $val = Cognition::RDFModel::Value->new($v, 1);
		$subject->addVal($ns.'tel', $val);
	}

	# Embedded objects which may be represented as resources
	foreach my $property (@_OBJECTS)
	{
		foreach my $V (@{ $this->{$property} })
		{
			my $val;
			if (ref $V)
				{ $val = Cognition::RDFModel::Value->new($V->uri, 1); }
			else
				{ $val = Cognition::RDFModel::Value->new($V); }
			$subject->addVal($ns.$property, $val);
		}
	}
	
	# TODO: 'agent' needs to be handled more intelligently. i.e. cope with raw strings
	# TODO: handle organisations
	
	return $subject;
	
} #/sub rdf_subject


sub metadata_ns
{
	my $this = shift;
	return { 'VCARD' => { nsurl=>'urn:ietf:rfc:2426#',title=>'vCard' } };
} #/sub namespace


sub parse
{
	my $page = shift;
	my $rv = { '_dom'=>shift };
	my $hcard_node = $rv->{'_dom'}->cloneNode(1);
	
	my $id = $hcard_node->getAttribute('id');
	$rv->{_id} = $id if (length $id);
	$rv->{_page} = $page;
	
	# Expand microformat patterns.
	my $incl_iterations = 0;
	my $replacements = 1;
	while (($incl_iterations < MAX_INCLUDE_ITERATIONS) && $replacements)
	{
		$replacements = Cognition::uF::ufp_include($hcard_node, $page) 
		              + Cognition::uF::ufp_include2($hcard_node, $page);
		$incl_iterations++;
	}
	Cognition::uF::ufp_abbr($hcard_node, $page);
	Cognition::uF::ufp_table($hcard_node, $page);

	# Check nested hcards to see if they have an 'agent' class, and if so, parse them as an agent.
	my @nested = Cognition::HTMLParser::Utils::searchClass('vcard', $hcard_node);
	foreach my $h (@nested)
	{
		if ($h->getAttribute('class') =~ / (^|\s) agent (\s|$) /x)
		{
			push @{ $rv->{agent} }, parse($page, $h);
		}
	}
	
	# Categories
	my @tags = Cognition::uF::RelTag::parse_all($page, $hcard_node);
	$rv->{category} = \@tags unless (!@tags);
	
	# Pull out embedded addresses.
	my @adrs = Cognition::uF::adr::parse_all($page, $hcard_node);
	$rv->{adr} = \@adrs unless (!@adrs);

	# Pull out embedded locations.
	my @geos = Cognition::uF::geo::parse_all($page, $hcard_node);
	$rv->{geo} = \@geos unless (!@geos);

	# Now that we have reached here, no other composite microformats have
	# any business being nested within this root element. So in the interest
	# of Microformat opacity, let's destroy them.
	Cognition::uF::destroyer($hcard_node);
	
	# Parse organisations.
	my @orgs = Cognition::HTMLParser::Utils::searchClass('org', $hcard_node);
	foreach my $org (@orgs)
	{
		my @orgnames = Cognition::HTMLParser::Utils::searchClass('organization-name', $org);
		my @orgunits = Cognition::HTMLParser::Utils::searchClass('organization-unit', $org);
		my $parsed_org = {};
		
		if (!@orgnames)
		{
			$parsed_org = { 
				'organisation-unit' => undef,
				'organisation-name' => Cognition::HTMLParser::Utils::STRINGIFY($org)
			};
		}
		else
		{
			my $orgname = $orgnames[0];
			my $orgunit = $orgunits[0];
			$parsed_org = { 
				'organisation-unit' => Cognition::HTMLParser::Utils::STRINGIFY($orgname),
				'organisation-name' => Cognition::HTMLParser::Utils::STRINGIFY($orgunit)
			};
		}
		
		push @{ $rv->{org} }, $parsed_org;
	}
	
	# Simple attributes - singles
	my @simples = qw(fn bday tz sort-string uid class family-name given-name
		additional-name honorific-prefix honorific-suffix);	
	foreach my $simple (@simples)
	{
		my @value_nodes = Cognition::HTMLParser::Utils::searchClass($simple, $hcard_node);
		my $value_node  = $value_nodes[0];
		next unless ($value_node);
		
		my $value = Cognition::HTMLParser::Utils::STRINGIFY($value_node);
		
		if ($simple eq 'uid')
		{
			if (length $value_node->getAttribute('id'))
				{ $value = $page->abs_url . '#' . $value_node->getAttribute('id'); }
			elsif (($value_node->localname eq 'a' || $value_node->localname eq 'area') && length $value_node->getAttribute('name'))
				{ $value = $page->abs_url . '#' . $value_node->getAttribute('name'); }
			elsif ($value_node->localname eq 'a' || $value_node->localname eq 'area')
				{ $value = $page->abs_url($value_node->getAttribute('href')); }
			elsif ($value_node->localname eq 'img')
				{ $value = $page->abs_url($value_node->getAttribute('src')); }
			elsif ($value_node->localname eq 'object')
				{ $value = $page->abs_url($value_node->getAttribute('data')); }
		}
		
		$rv->{$simple} = $value;
	}
	
	if (length $rv->{bday} && length $rv->{tz})
		{ $rv->{bday} = Cognition::uF::datetime($rv->{bday}, $rv->{tz}); }
	elsif (length $rv->{bday})
		{ $rv->{bday} = Cognition::uF::datetime($rv->{bday}); }
	
	# Simple attributes - possible duplicates
	my @simples = qw(nickname url label title role photo sound key logo mailer
		category note rev agent);	
	foreach my $simple (@simples)
	{
		my @value_nodes = Cognition::HTMLParser::Utils::searchClass($simple, $hcard_node);
		foreach my $value_node (@value_nodes)
		{
			next unless ($value_node);
			
			my $value = Cognition::HTMLParser::Utils::STRINGIFY($value_node);
			
			if ($simple =~ /^(url|photo|logo|sound|key)$/)
			{
				if ($value_node->localname eq 'a' || $value_node->localname eq 'area')
					{ $value = $page->abs_url($value_node->getAttribute('href')); }
				elsif ($value_node->localname eq 'img')
					{ $value = $page->abs_url($value_node->getAttribute('src')); }
				elsif ($value_node->localname eq 'object')
					{ $value = $page->abs_url($value_node->getAttribute('data')); }
			}

			if ($simple eq 'rev' && length $value)
				{ $value = Cognition::uF::datetime($value); }
			
			push @{$rv->{$simple}}, $value;
		}
	}
	
	# Telephone numbers and e-mail addresses
	my @comms = ('tel', 'email');
	foreach my $comm (@comms)
	{
		my @nodes = Cognition::HTMLParser::Utils::searchClass($comm, $hcard_node);
		foreach my $n (@nodes)
		{
			my @accumValue;
			my @accumType;
			
			foreach my $v (Cognition::HTMLParser::Utils::searchClass('value', $n))
			{
				if ($v->localname eq 'a' || $v->localname eq 'area')
				{
					my $href = $v->getAttribute('href');
					$href =~ s/(^[a-z]+\:|\?.*$)//g;
					push @accumValue, $href;
				}
				elsif ($v)
				{
					push @accumValue, Cognition::HTMLParser::Utils::STRINGIFY($v);
				}
			}
			
			foreach my $v (Cognition::HTMLParser::Utils::searchClass('type', $n))
			{
				push @accumType, uc(Cognition::HTMLParser::Utils::STRINGIFY($v));
				$v->getParentNode->removeChild($v);
			}
			
			my $val = join '', @accumValue;
			if ($val !~ /[0-9a-z]/i)
			{
				$val = Cognition::HTMLParser::Utils::STRINGIFY($n);
			}
			if ($comm eq 'tel')
			{
				$val = lc($val);
				$val =~ s/[^0-9x+-]//g;
			}
			
			push @{ $rv->{$comm} }, {
					'value' => $val,
					'type'  => \@accumType
				};
		}
		
	}
	
	# Check to see if this contact card is for an individual contact or an org.
	# If for an org, disable n-optimisation.
	my $do_n_opt = 1;
	$rv->{_type} = 'contact';
	foreach my $org (@{ $rv->{org} })
	{
		if (lc($org->{'organisation-name'}) eq lc($rv->{fn}))
		{
			$rv->{_type} = 'organisation';
			$do_n_opt = 0;
			last;
		}
	}

	# Perform n-optimisation.
	if ($do_n_opt)
	{
		foreach my $f (qw(family-name given-name additional-name honorific-prefix honorific-suffix))
		{
			if (length $rv->{$f})
			{
				$do_n_opt = 0;
			}
		}
		
		if ($do_n_opt)
		{
			$rv->{fn} =~ s/(^\s|\s$)//g;
			$rv->{fn} =~ s/\s+/ /g;
			
			my @words = split / /, $rv->{fn};
			
			if (scalar @words == 1)
			{
				$rv->{nickname} = [ $words[0] ];
			}
			elsif (scalar @words)
			{
				if ($words[0] =~ /^.*\,$/ || $words[1] =~ /^.\.?$/)
				{
					$words[0] =~ s/[\.\,]$//;
					$words[1] =~ s/[\.\,]$//;
					$rv->{'given-name'} = $words[1];
					$rv->{'family-name'} = $words[0];
				}
				else
				{
					$rv->{'given-name'} = $words[0];
					$rv->{'family-name'} = $words[-1];
				}
			}
		}
	}
	
	unless (defined $rv->{n})
	{
		my $additional = $rv->{'additional-name'};
		$additional =~ s/\s/\,/g;
		$rv->{n} = join ';', $rv->{'family-name'}, $rv->{'given-name'}, $additional,
		                     $rv->{'honorific-prefix'}, $rv->{'honorific-suffix'};
	}
	
	bless $rv;
	return $rv;
	
} #/sub parse


1; #/package Cognition::uF::hCard




